from __future__ import annotations

import datetime
import warnings
from typing import TYPE_CHECKING, Union, cast

import dateutil.parser
import isodate

from chalk.features._encoding.missing_value import MissingValueStrategy
from chalk.features.feature_field import Feature

if TYPE_CHECKING:
    import polars as pl

try:
    import zoneinfo
except ImportError:
    # Zoneinfo was introduced in python 3.9
    from backports import zoneinfo


def _polars_dtype_contains_struct(dtype: pl.PolarsDataType):
    """Returns whether the dtype contains a (potentially nested) struct"""
    import polars as pl

    if isinstance(dtype, pl.Struct) or (isinstance(dtype, type) and issubclass(dtype, pl.Struct)):
        return True
    if isinstance(dtype, pl.List):
        assert dtype.inner is not None
        return _polars_dtype_contains_struct(dtype.inner)
    return False


def _generate_empty_series_for_dtype(name: str, dtype: pl.PolarsDataType, length: int) -> pl.Series:
    """Safely generate a series of all null values for the specified datatype.

    Unlike the ``pl.Series`` constructor, this function can handle struct dtypes.
    """
    import polars as pl

    if isinstance(dtype, pl.Struct):
        # Struct dtypes cannot be specified in the pl.Series constructor.
        # Instead, create a dataframe, then call .to_struct() on it
        # If recursing within a struct, it should have a length of zero
        data = {f.name: _generate_empty_series_for_dtype(f.name, f.dtype, length) for f in dtype.fields}
        temp_df = pl.DataFrame(data)
        return temp_df.to_struct(name)
    if isinstance(dtype, pl.List) and isinstance(dtype.inner, pl.Struct):
        # The pl.Series constructor does not respect nested data types
        # So, we'll manually build a dataframe with the list set to the correct type,
        # set it to None, and then select just the list
        assert dtype.inner is not None
        data = {name: _generate_empty_series_for_dtype(name, dtype.inner, 0)}
        temp_df = pl.DataFrame(data)
        df_with_list = temp_df.select(pl.col(name).reshape((length, -1)))
        df_with_list = df_with_list.select(pl.when(True).then(None).otherwise(pl.col(name)).alias(name))
        list_of_struct_series = df_with_list.select(pl.col(name).reshape((length, -1))).get_column(name)
        return list_of_struct_series
    return pl.Series(name, dtype=dtype, values=([None] * length))


def _iso_parse(x: str, expected_tz: datetime.tzinfo | None) -> datetime.datetime:
    """This function converts the input into an expected timezone, and drops it, so we can cast it in polars with with_time_zone"""
    parsed = dateutil.parser.parse(x)
    if parsed.tzinfo is not None and expected_tz is not None:
        # Convert the timezone
        parsed = parsed.astimezone(expected_tz)
    # Now drop the timezone
    return parsed.replace(tzinfo=None)


def validate_df_schema(underlying: Union[pl.DataFrame, pl.LazyFrame]):
    # This is called from within DataFrame.__init__, which validates that polars is installed
    import polars as pl

    for root_fqn, actual_dtype in underlying.schema.items():
        feature = Feature.from_root_fqn(root_fqn)
        if feature.is_has_one or feature.is_has_many:
            continue
        expected_dtype = feature.converter.polars_dtype
        if actual_dtype == expected_dtype:
            continue
        if isinstance(underlying, pl.LazyFrame):
            underlying = underlying.collect()
        if len(underlying) == underlying.get_column(root_fqn).null_count():
            # If all values are null, then replace with a null series of the correct datatype
            # It's quite possible that the original column will have an incorrect dtype if all values are null,
            # since there was no data to infer the correct dtype from
            underlying = underlying.with_column(
                _generate_empty_series_for_dtype(root_fqn, expected_dtype, len(underlying))
            )
        elif _polars_dtype_contains_struct(expected_dtype):
            # Cannot cast to a struct type. Instead, will error, so the user can ensure the underlying
            # dictionaries / dataclasses are of the correct type
            raise TypeError(f"Expected field '{root_fqn}' to have dtype `{expected_dtype}`; got dtype `{actual_dtype}`")
        else:
            if actual_dtype == pl.Utf8:
                if isinstance(expected_dtype, pl.Datetime):
                    tzinfo = None if expected_dtype.tz is None else zoneinfo.ZoneInfo(expected_dtype.tz)
                    cast_expr = (
                        pl.col(root_fqn)
                        # Ignoring the timezone here because we set it after the parse
                        .apply(
                            lambda x: None if x is None else _iso_parse(cast(str, x), tzinfo),
                        )
                        .cast(pl.Datetime(expected_dtype.tu, None))
                        .dt.with_time_zone(expected_dtype.tz)
                    )
                elif expected_dtype == pl.Date:
                    cast_expr = pl.col(root_fqn).apply(
                        lambda x: None if x is None else isodate.parse_date(x),
                    )
                elif expected_dtype == pl.Time:
                    cast_expr = pl.col(root_fqn).apply(
                        lambda x: None if x is None else isodate.parse_time(x),
                    )
                elif expected_dtype == pl.Duration:
                    cast_expr = pl.col(root_fqn).apply(
                        lambda x: None if x is None else isodate.parse_duration(x),
                    )
                else:
                    cast_expr = pl.col(root_fqn).cast(expected_dtype)
                col = cast_expr.alias(root_fqn)
            else:
                col = pl.col(root_fqn).cast(expected_dtype)
            try:
                underlying = underlying.with_column(col)
            except pl.ComputeError as e:
                raise TypeError(
                    f"Values for feature `{root_fqn}` could not be converted to dtype `{expected_dtype.string_repr()}`. Found type {actual_dtype}, instead."
                ) from e
    return underlying


def validate_nulls(
    underlying: Union[pl.DataFrame, pl.LazyFrame],
    missing_value_strategy: MissingValueStrategy,
) -> pl.DataFrame:
    """Validate that any null values are in columns that support nullable values"""
    # This is called from within DataFrame.__init__, which validates that polars is installed
    import polars as pl

    if isinstance(underlying, pl.LazyFrame):
        underlying = underlying.collect()
    schema = underlying.schema
    null_count_rows = underlying.null_count().to_dicts()
    if len(null_count_rows) == 0:
        return underlying  # Empty dataframe
    assert len(null_count_rows) == 1
    null_counts = null_count_rows[0]
    for col_name, null_count in null_counts.items():
        feature = Feature.from_root_fqn(col_name)
        if null_count > 0 and not feature.typ.is_nullable and not isinstance(schema[col_name], pl.Struct):
            if missing_value_strategy == "allow":
                warnings.warn(
                    UserWarning(f"Allowing missing value for feature '{col_name}' with strategy 'default_or_allow'")
                )
            elif missing_value_strategy == "error":
                raise TypeError(f"Feature '{col_name}' has missing values, but the feature is non-nullable")
            elif missing_value_strategy in ("default_or_error", "default_or_allow"):
                if feature.converter.has_default:
                    underlying = underlying.with_columns(
                        [pl.col(col_name).fill_null(feature.converter.primitive_default)]
                    )
                elif missing_value_strategy == "default_or_error":
                    raise TypeError(f"Feature '{col_name}' has missing values, but the feature does not have a default")
                else:
                    warnings.warn(
                        UserWarning(f"Allowing missing value for feature '{col_name}' with strategy 'default_or_allow'")
                    )
            else:
                raise ValueError(
                    (
                        f"Unsupported missing value strategy '{missing_value_strategy}'. "
                        "Allowed options are 'allow', 'error', and 'default_or_error'."
                    )
                )
    return underlying



from __future__ import annotations

import collections.abc
import dataclasses
from datetime import timedelta
from typing import TYPE_CHECKING, Any, Set

if TYPE_CHECKING:
    from chalk.features.feature_field import Feature


__all__ = ["Filter", "before", "after", "TimeDelta"]


class Filter:
    def __init__(self, lhs: Any, operation: str, rhs: Any):
        from chalk.features.feature_wrapper import FeatureWrapper, unwrap_feature

        # Feature or other could be another feature, filter, featuretime, literal
        # Other could also be a sequence (in the case of operation = "in")
        self.operation = operation
        if isinstance(lhs, FeatureWrapper):
            lhs = unwrap_feature(lhs)
        self.lhs = lhs
        if self.operation == "in":
            if not isinstance(rhs, collections.abc.Iterable):
                raise ValueError("The RHS must be an iterable for operation='in'")
            rhs = {unwrap_feature(x) if isinstance(x, FeatureWrapper) else x for x in rhs}
        else:
            if isinstance(rhs, FeatureWrapper):
                rhs = unwrap_feature(rhs)
        self.rhs = rhs

    def __hash__(self) -> int:
        return hash((self.lhs, self.operation, self.rhs))

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Filter):
            return NotImplemented
        return self.lhs == other.lhs and self.operation == other.operation and self.rhs == other.rhs

    def __and__(self, other: object):
        return Filter(self, "and", other)

    def __or__(self, other: object):
        return Filter(self, "or", other)

    def __invert__(self):
        return Filter(self, "not", None)

    def __bool__(self):
        # If the operation is == or !=, then compare the underlying filters or features
        # If x and y are features, then (x == y) is overloaded to return a Filter
        # This allows bool(x == y) to return True iff the x is y, while (x == y) is still a Filter instance
        # If either side is a FeatureWrapper
        # If either side is a Filter or a FeatureWrapper,
        if self.operation == "==":
            return self.lhs == self.rhs
        if self.operation == "!=":
            return self.lhs != self.rhs
        # Non-equality comparisons must be evaluated manually -- e.g. if x and y are Features
        # (x < y) would return a Filter, but bool(x < y) doesn't really make sense. So, raise.
        raise TypeError(f"Operation {self.operation} on a Filter is undefined")

    def referenced_features(self) -> Set[Feature]:
        from chalk.features.feature_field import Feature

        return {x for x in (self.lhs, self.rhs) if isinstance(x, Feature)}

    def __repr__(self):
        return f"Filter({self.lhs} {self.operation} {self.rhs})"


@dataclasses.dataclass(frozen=True)
class TimeDelta:
    weeks_ago: int = 0
    days_ago: int = 0
    hours_ago: int = 0
    minutes_ago: int = 0
    seconds_ago: int = 0
    milliseconds_ago: int = 0
    microseconds_ago: int = 0

    def to_std(self) -> timedelta:
        # Returning the negative inverse since the feature is defined in the past (e.g. days **ago**)
        return -timedelta(
            weeks=self.weeks_ago,
            days=self.days_ago,
            hours=self.hours_ago,
            minutes=self.minutes_ago,
            seconds=self.seconds_ago,
            milliseconds=self.milliseconds_ago,
            microseconds=self.microseconds_ago,
        )


def before(
    weeks_ago: int = 0,
    days_ago: int = 0,
    hours_ago: int = 0,
    minutes_ago: int = 0,
    seconds_ago: int = 0,
    milliseconds_ago: int = 0,
    microseconds_ago: int = 0,
) -> Any:
    """The function `before` can be used with `DataFrame` to compute windowed features.

    Before filters a `DataFrame` relative to the current time in context.
    This time could be in the past if you’re using an offline resolver.
    Using window functions ensures that you maintain point-in-time correctness.

    The parameters to `before` take many keyword arguments describing the
    time relative to the present.

    Parameters
    ----------
    days_ago
        Number of days ago.
    hours_ago
        Number of hours ago.
    minutes_ago
        Number of minutes ago.
    seconds_ago
        Number of seconds ago.

    Other Parameters
    ----------------
    weeks_ago
        Number of weeks ago.
    milliseconds_ago
        Number of milliseconds ago.
    microseconds_ago
        Number of microseconds ago.

    Returns
    -------
    Any
        A filter for a `DataFrame`.

    Examples
    --------
    >>> from chalk.features import DataFrame, features
    >>> @features
    ... class Card:
    ...     ...
    >>> @features
    ... class User:
    ...     cards: DataFrame[Card]
    >>> User.cards[before(hours_ago=1, minutes_ago=30)]
    """
    from chalk.features.pseudofeatures import CHALK_TS_FEATURE

    return Filter(
        lhs=CHALK_TS_FEATURE,
        operation="<=",
        rhs=TimeDelta(
            weeks_ago=weeks_ago,
            days_ago=days_ago,
            hours_ago=hours_ago,
            minutes_ago=minutes_ago,
            seconds_ago=seconds_ago,
            milliseconds_ago=milliseconds_ago,
            microseconds_ago=microseconds_ago,
        ),
    )


def after(
    weeks_ago: int = 0,
    days_ago: int = 0,
    hours_ago: int = 0,
    minutes_ago: int = 0,
    seconds_ago: int = 0,
    milliseconds_ago: int = 0,
    microseconds_ago: int = 0,
) -> Any:
    """The function `after` can be used with `DataFrame` to compute windowed features.

    After filters a `DataFrame` relative to the current time in context.
    This time could be in the past if you’re using an offline resolver.
    Using window functions ensures that you maintain point-in-time correctness.

    The parameters to `after` take many keyword arguments describing the
    time relative to the present.

    Parameters
    ----------
    days_ago
        Number of days ago.
    hours_ago
        Number of hours ago.
    minutes_ago
        Number of minutes ago.
    seconds_ago
        Number of seconds ago.

    Other Parameters
    ----------------
    weeks_ago
        Number of weeks ago.
    milliseconds_ago
        Number of milliseconds ago.
    microseconds_ago
        Number of microseconds ago.

    Returns
    -------
    Any
        A filter for the `DataFrame`.

    Examples
    --------
    >>> from chalk.features import DataFrame, features
    >>> @features
    ... class Card:
    ...     ...
    >>> @features
    ... class User:
    ...     cards: DataFrame[Card]
    >>> User.cards[after(hours_ago=1, minutes_ago=30)]
    """
    from chalk.features.pseudofeatures import CHALK_TS_FEATURE

    return Filter(
        lhs=CHALK_TS_FEATURE,
        operation=">=",
        rhs=TimeDelta(
            weeks_ago=weeks_ago,
            days_ago=days_ago,
            hours_ago=hours_ago,
            minutes_ago=minutes_ago,
            seconds_ago=seconds_ago,
            milliseconds_ago=milliseconds_ago,
            microseconds_ago=microseconds_ago,
        ),
    )



# AUTO-GENERATED FILE.
# Re-run /chalk/feature_n/codegen.py to regenerate contents.
from typing import TypeVar, Generic, Optional, Dict
from chalk.features.dataframe import DataFrameMeta


T1 = TypeVar("T1")
T2 = TypeVar("T2")
T3 = TypeVar("T3")
T4 = TypeVar("T4")
T5 = TypeVar("T5")
T6 = TypeVar("T6")
T7 = TypeVar("T7")
T8 = TypeVar("T8")


class Features(Generic[T1, T2, T3, T4, T5, T6, T7, T8]):
    pass


class DataFrame(Generic[T1, T2, T3, T4, T5, T6, T7, T8], metaclass=DataFrameMeta):
    def __getitem__(self, item):
        pass



from chalk.features.dataframe._filters import convert_filters_to_pl_expr
from chalk.features.dataframe._impl import DataFrame, DataFrameMeta

__all__ = ["DataFrame", "DataFrameMeta", "convert_filters_to_pl_expr"]



import dataclasses
import zlib
from copy import deepcopy
from typing import Any, Dict, List, Literal, Optional, Tuple, TypeVar, Union

from chalk._monitoring.charts_enums_codegen import (
    ChartLinkKind,
    ComparatorKind,
    FilterKind,
    GroupByKind,
    MetricKind,
    WindowFunctionKind,
)
from chalk.features import FeatureWrapper
from chalk.features.resolver import Resolver, ResolverProtocol


@dataclasses.dataclass
class MetricFilter:
    kind: FilterKind
    comparator: ComparatorKind
    value: List[str]


@dataclasses.dataclass
class ThresholdFunction:
    lhs: "SeriesBase"
    operation: str
    rhs: float


ResolverType = Literal["online", "offline", "stream"]
TSeries = TypeVar("TSeries", bound="SeriesBase")


class SeriesBase:
    """
    Base class for Series
    """

    def __init__(
        self,
        name: str,
        metric: Union[MetricKind, str],
        window_function: Optional[Union[WindowFunctionKind, str]] = None,
        time_shift: Optional[int] = None,
    ):
        self._name = name
        self._metric = MetricKind(metric.upper()) if metric else None
        self._filters: List[MetricFilter] = []
        if window_function:
            if WindowFunctionKind.has_member_key(window_function.upper()):
                window_function_kind = WindowFunctionKind(window_function.upper())
            else:
                window_function_kind = self._get_window_function_type(window_function.upper())
                if not window_function_kind:
                    raise ValueError(f"'window_function' value '{window_function}' 'invalid for WindowFunctionKind")
        else:
            window_function_kind = None
        self._window_function = window_function_kind
        self._group_by: List[GroupByKind] = []
        self._time_shift = time_shift
        self._entity_kind = ChartLinkKind.manual
        self._entity_id = None

    @staticmethod
    def _get_window_function_type(key: str) -> Union[WindowFunctionKind, None]:
        window_function_value_dict = {
            "99%": WindowFunctionKind.PERCENTILE_99,
            "95%": WindowFunctionKind.PERCENTILE_95,
            "75%": WindowFunctionKind.PERCENTILE_75,
            "50%": WindowFunctionKind.PERCENTILE_50,
            "25%": WindowFunctionKind.PERCENTILE_25,
            "5%": WindowFunctionKind.PERCENTILE_5,
            "99": WindowFunctionKind.PERCENTILE_99,
            "95": WindowFunctionKind.PERCENTILE_95,
            "75": WindowFunctionKind.PERCENTILE_75,
            "50": WindowFunctionKind.PERCENTILE_50,
            "25": WindowFunctionKind.PERCENTILE_25,
            "5": WindowFunctionKind.PERCENTILE_5,
            "ALL": WindowFunctionKind.ALL_PERCENTILES,
        }
        return window_function_value_dict.get(key)

    def _where(
        self: TSeries,
        feature: Optional[Union[List[Any], Any]] = None,
        resolver: Optional[Union[List[Union[ResolverProtocol, str]], Union[ResolverProtocol, str]]] = None,
        feature_tag: Optional[Union[List[str], str]] = None,
        resolver_tag: Optional[Union[List[str], str]] = None,
        operation_id: Optional[Union[List[str], str]] = None,
        query_name: Optional[Union[List[str], str]] = None,
        feature_status: Optional[Literal["success", "failure"]] = None,
        resolver_status: Optional[Literal["success", "failure"]] = None,
        cron_status: Optional[Literal["success", "failure"]] = None,
        migration_status: Optional[Literal["success", "failure"]] = None,
        query_status: Optional[Literal["success", "failure"]] = None,
        resolver_type: Optional[Union[List[ResolverType], ResolverType]] = None,
        cache_hit: Optional[bool] = None,
        is_null: Optional[bool] = None,
        equals: Optional[bool] = True,
    ) -> TSeries:
        copy = self._copy_with()
        success_dict = {"success": True, "failure": False}
        if feature:
            feature_name = [feature] if not isinstance(feature, list) else feature
            copy = copy._feature_name_filter(*feature_name, equals=equals)
        if resolver:
            resolver_name = [resolver] if not isinstance(resolver, list) else resolver
            copy = copy._resolver_name_filter(*resolver_name, equals=equals)
        if feature_tag:
            feature_tag = [feature_tag] if isinstance(feature_tag, str) else feature_tag
            copy = copy._string_filter(*feature_tag, kind=FilterKind.FEATURE_TAG, equals=equals)
        if resolver_tag:
            resolver_tag = [resolver_tag] if isinstance(resolver_tag, str) else resolver_tag
            copy = copy._string_filter(*resolver_tag, kind=FilterKind.RESOLVER_TAG, equals=equals)
        if operation_id:
            operation_id = [operation_id] if isinstance(operation_id, str) else operation_id
            copy = copy._string_filter(*operation_id, kind=FilterKind.OPERATION_ID, equals=equals)
        if query_name:
            query_name = [query_name] if isinstance(query_name, str) else query_name
            copy = copy._string_filter(*query_name, kind=FilterKind.QUERY_NAME, equals=equals)
        if feature_status:
            copy = copy._status_filter(kind=FilterKind.FEATURE_STATUS, success=success_dict[feature_status] == equals)
        if resolver_status:
            copy = copy._status_filter(kind=FilterKind.RESOLVER_STATUS, success=success_dict[resolver_status] == equals)
        if cron_status:
            copy = copy._status_filter(kind=FilterKind.CRON_STATUS, success=success_dict[cron_status] == equals)
        if migration_status:
            copy = copy._status_filter(
                kind=FilterKind.MIGRATION_STATUS, success=success_dict[migration_status] == equals
            )
        if query_status:
            copy = copy._status_filter(kind=FilterKind.QUERY_STATUS, success=success_dict[query_status] == equals)
        if resolver_type:
            resolver_type = [resolver_type] if not isinstance(resolver_type, list) else resolver_type
            copy = copy._with_resolver_type_filter(*resolver_type, equals=equals)
        if cache_hit is not None:
            copy = copy._true_false_filter(kind=FilterKind.CACHE_HIT, value=cache_hit == equals)
        if is_null is not None:
            copy = copy._true_false_filter(kind=FilterKind.IS_NULL, value=is_null == equals)
        return copy

    def _feature_name_filter(self: TSeries, *features: Tuple[Any], equals: bool) -> TSeries:
        if not features:
            raise ValueError(f"One or more Chalk Features must be supplied.")
        copy = self._copy_with()
        comparator = ComparatorKind.EQ if equals else ComparatorKind.NEQ
        if len(features) == 1 or not equals:
            for feature in features:
                value = str(feature)
                metric_filter = MetricFilter(kind=FilterKind.FEATURE_NAME, comparator=comparator, value=[value])
                copy._filters.append(metric_filter)
            if len(features) == 1:
                feature = features[0]
                copy._entity_id = str(feature)
                copy._entity_kind = ChartLinkKind.feature
        else:
            value = [str(feature) for feature in features]
            metric_filter = MetricFilter(kind=FilterKind.FEATURE_NAME, comparator=ComparatorKind.ONE_OF, value=value)
            copy._filters.append(metric_filter)
        return copy

    def _resolver_name_filter(self: TSeries, *resolvers: Tuple[Union[ResolverProtocol, str]], equals: bool) -> TSeries:
        if not resolvers:
            raise ValueError(f"One or more Chalk Resolvers must be supplied.")
        copy = self._copy_with()
        comparator = ComparatorKind.EQ if equals else ComparatorKind.NEQ
        if len(resolvers) == 1 or not equals:
            for resolver in resolvers:
                value = resolver.fqn if isinstance(resolver, Resolver) else resolver
                metric_filter = MetricFilter(kind=FilterKind.RESOLVER_NAME, comparator=comparator, value=[value])
                copy._filters.append(metric_filter)
            if len(resolvers) == 1:
                resolver = resolvers[0]
                copy._entity_id = resolver.fqn if isinstance(resolver, Resolver) else resolver
                copy._entity_kind = ChartLinkKind.resolver
        else:
            value = [resolver.fqn if isinstance(resolver, Resolver) else resolver for resolver in resolvers]
            metric_filter = MetricFilter(kind=FilterKind.RESOLVER_NAME, comparator=ComparatorKind.ONE_OF, value=value)
            copy._filters.append(metric_filter)
        return copy

    def _string_filter(self: TSeries, *strings: Tuple[str], kind: FilterKind, equals=True) -> TSeries:
        if not strings:
            raise ValueError(f"One or more arguments must be supplied for this filter")
        copy = self._copy_with()
        comparator = ComparatorKind.EQ if equals else ComparatorKind.NEQ
        if len(strings) == 1 or not equals:
            for string in strings:
                metric_filter = MetricFilter(kind=kind, comparator=comparator, value=[string])
                copy._filters.append(metric_filter)
            if len(strings) == 1 and kind == FilterKind.QUERY_NAME:
                copy._entity_id = strings[0]
                copy._entity_kind = ChartLinkKind.query
        else:
            metric_filter = MetricFilter(kind=kind, comparator=ComparatorKind.ONE_OF, value=list(strings))
            copy._filters.append(metric_filter)
        return copy

    def _with_resolver_type_filter(
        self: TSeries, *resolver_types: Tuple[Literal["online", "offline", "stream"]], equals=True
    ) -> TSeries:
        if not resolver_types:
            raise ValueError(f"One or more resolver types from 'online', 'offline', or 'stream' must be supplied")
        if not set(resolver_types).issubset(["online", "offline", "stream"]):
            raise ValueError(f"Resolver types '{resolver_types}' must be one of 'online', 'offline', or 'stream'")
        copy = self._copy_with()
        comparator = ComparatorKind.EQ if equals else ComparatorKind.NEQ
        if len(resolver_types) == 1 or not equals:
            for resolver_type in resolver_types:
                metric_filter = MetricFilter(
                    kind=FilterKind.ONLINE_OFFLINE, comparator=comparator, value=[resolver_type]
                )
                copy._filters.append(metric_filter)
        else:
            metric_filter = MetricFilter(
                kind=FilterKind.ONLINE_OFFLINE, comparator=ComparatorKind.ONE_OF, value=list(resolver_types)
            )
            copy._filters.append(metric_filter)
        return copy

    def _true_false_filter(self: TSeries, kind: FilterKind, value: bool) -> TSeries:
        copy = self._copy_with()
        value = "true" if value else "false"
        metric_filter = MetricFilter(kind=kind, comparator=ComparatorKind.EQ, value=[value])
        copy._filters.append(metric_filter)
        return copy

    def _status_filter(self: TSeries, kind: FilterKind, success: bool) -> TSeries:
        copy = self._copy_with()
        value = "success" if success else "failure"
        metric_filter = MetricFilter(kind=kind, comparator=ComparatorKind.EQ, value=[value])
        copy._filters.append(metric_filter)
        return copy

    def _with_filter(
        self: TSeries,
        kind: Union[FilterKind, str],
        comparator: Union[ComparatorKind, str],
        value: Union[List[str], str],
    ) -> TSeries:
        copy = self._copy_with()
        kind = FilterKind(kind.upper())
        comparator = ComparatorKind(comparator.upper())
        value = [value] if isinstance(value, str) else value
        metric_filter = MetricFilter(kind=kind, comparator=comparator, value=value)
        copy._filters.append(metric_filter)
        return copy

    def with_window_function(self: TSeries, window_function: Union[WindowFunctionKind, str]) -> TSeries:
        copy = self._copy_with()
        copy._window_function = WindowFunctionKind(window_function.upper())
        return copy

    def with_group_by(self: TSeries, group_by: Union[GroupByKind, str]) -> TSeries:
        copy = self._copy_with()
        group_by = GroupByKind(group_by.upper())
        copy._group_by.append(group_by)
        return copy

    def with_time_shift(self: TSeries, time_shift: int) -> TSeries:
        copy = self._copy_with()
        copy._time_shift = time_shift
        return copy

    def _copy_with(self: TSeries) -> TSeries:
        self_copy = deepcopy(self)
        return self_copy

    def __gt__(self, other) -> ThresholdFunction:
        return ThresholdFunction(self, ">", other)

    def __lt__(self, other) -> ThresholdFunction:
        return ThresholdFunction(self, "<", other)

    def __str__(self) -> str:
        return f"Series(name='{self._name}')"

    def __hash__(self) -> int:
        name = self._name if self._name else "."
        metric = str(self._metric) if self._metric else "."
        filter_strings = (
            sorted([f"{f.kind}.{f.comparator}.{'.'.join(f.value)}" for f in self._filters]) if self._filters else "."
        )
        window_function = str(self._window_function) if self._window_function else "."
        group_by = sorted([str(group_by) for group_by in self._group_by]) if self._group_by else "."
        time_shift = str(self._time_shift) if self._time_shift else "."

        series_string = (
            f"series.{name}.{metric}.{'.'.join(filter_strings)}.{window_function}.{'.'.join(group_by)}.{time_shift}"
        )

        return zlib.crc32(series_string.encode())



from datetime import datetime
from typing import Any

from typing_extensions import Annotated

from chalk.features.feature_field import Feature
from chalk.features.feature_wrapper import unwrap_feature

FeatureTime = Annotated[datetime, "__chalk_ts__"]
"""Marker for a FeatureTime.

See https://docs.chalk.ai/docs/time for more details on feature time.

Examples
--------
>>> from chalk.features import features
>>> @features
... class User:
...     id: str
...     updated_at: FeatureTime
"""


def feature_time() -> Any:
    """Declare a feature time (deprecated).
    Prefer to use the `FeatureTime` marker.
    See https://docs.chalk.ai/docs/time for more details on `FeatureTime`.

    Examples
    --------
    >>> from chalk.features import features
    >>> @features
    ... class User:
    ...     updated_at: datetime = feature_time()
    """
    return Feature(typ=datetime, is_feature_time=True)


def is_feature_time(f: Any) -> bool:
    """Determine whether a feature is a feature time.
    See https://docs.chalk.ai/docs/time for more details on `FeatureTime`.

    Parameters
    ----------
    f
        A feature (i.e. `User.ts`)

    Returns
    -------
    bool
        True if the feature is a `FeatureTime` and `False` otherwise

    Examples
    --------
    >>> from chalk.features import features
    >>> @features
    ... class User:
    ...     id: str
    ...     updated_at: datetime = feature_time()
    >>> assert is_feature_time(User.updated_at) is True
    >>> assert is_feature_time(User.id) is False
    """
    return unwrap_feature(f).is_feature_time



import os
import shutil
from pathlib import Path


def generate_feature_n(n: int, preamble: str) -> str:
    typevars = "\n".join([f"T{str(i)} = TypeVar('T{str(i)}')" for i in range(1, n + 1)])

    all_types = [f"T{str(i)}" for i in range(1, n + 1)]
    all_types_str = ", ".join(all_types)

    cls = f"""


class Features(Generic[{all_types_str}]):
    pass


class DataFrame(Generic[{all_types_str}], metaclass=DataFrameMeta):
    def __getitem__(self, item):
        pass
"""

    return f"{preamble}{typevars}{cls}"


def codegen(num_classes: int):
    dir_path = os.path.dirname(os.path.realpath(__file__))
    codegen_filename = f"{dir_path}/codegen.py".split("cli", 1)[1]
    preamble = f"""# AUTO-GENERATED FILE.
# Re-run {codegen_filename} to regenerate contents.

"""
    for i in range(1, num_classes + 1):
        file_content = generate_feature_n(i, preamble)
        feature_n_dir = f"{dir_path}/feature_{str(i)}/"
        shutil.rmtree(feature_n_dir, ignore_errors=True)
        Path(feature_n_dir).mkdir()
        Path(f"{feature_n_dir}/__init__.py").touch()
        Path(f"{feature_n_dir}/feature.py").write_text(file_content)

    init_text = "\n".join(
        [f"from .feature_{str(i)}.feature import Features as Feature{str(i)}" for i in range(1, num_classes + 1)]
    )
    Path(f"{dir_path}/__init__.py").write_text(f"{preamble}{init_text}")


if __name__ == "__main__":
    codegen(num_classes=256)



import os
from functools import lru_cache
from pathlib import Path
from typing import Mapping, Optional

import yaml
from pydantic import BaseModel

from chalk.config.project_config import load_project_config
from chalk.utils.log_with_context import get_logger

_logger = get_logger(__name__)


class TokenConfig(BaseModel):
    name: Optional[str] = None
    clientId: str
    clientSecret: str
    apiServer: Optional[str] = None
    activeEnvironment: Optional[str] = None


class AuthConfig(BaseModel):
    tokens: Optional[Mapping[str, TokenConfig]]


def _load_global_config_uncached():
    home = os.getenv("XDG_CONFIG_HOME", os.path.expanduser("~"))
    path = None
    for filename in [".chalk.yml"]:
        p = os.path.join(home, filename)
        if os.path.exists(p):
            path = p

    if path is None:
        return None

    with open(path, "r") as f:
        parsed = yaml.safe_load(f)
        return AuthConfig(**parsed)


# using lru_cache for 3.8 compat
@lru_cache(maxsize=None)
def _load_global_config() -> Optional[AuthConfig]:
    return _load_global_config_uncached()


def _en(current_env_name: str, old_env_name: Optional[str] = None) -> Optional[str]:
    old_env_name = old_env_name or f"_{current_env_name}"
    # Backwards compatability for old env vars.
    deprecated = os.getenv(old_env_name)
    current = os.getenv(current_env_name)
    if deprecated and current:
        _logger.warning(
            f"Using both deprecated '{old_env_name}' and new '{current_env_name}' environment variables. Choosing '{current_env_name}'"
        )
    return current or deprecated


def load_token(
    client_id: Optional[str],
    client_secret: Optional[str],
    active_environment: Optional[str],
    api_server: Optional[str],
    skip_cache: bool = False,
) -> Optional[TokenConfig]:
    token_name = "Explicit" if client_id and client_secret else "Environment"

    client_id = client_id or _en("CHALK_CLIENT_ID")
    client_secret = client_secret or _en("CHALK_CLIENT_SECRET")
    api_server = api_server or _en("CHALK_API_SERVER")
    active_environment = active_environment or _en("CHALK_ENVIRONMENT", "_CHALK_ACTIVE_ENVIRONMENT")

    if client_id is not None and client_secret is not None:
        return TokenConfig(
            name=token_name,
            clientId=client_id,
            clientSecret=client_secret,
            apiServer=api_server or "https://api.chalk.ai",
            activeEnvironment=active_environment,
        )

    global_cfg = _load_global_config_uncached() if skip_cache else _load_global_config()
    if global_cfg is None:
        return None

    absdir = Path(os.getcwd()).absolute().resolve()
    project_config = load_project_config()
    if project_config:
        absdir = Path(project_config.local_path).parent
    tokens: Mapping[str, TokenConfig] = global_cfg.tokens or {}
    token_config = tokens.get(str(absdir)) or tokens.get("default")
    if token_config is None:
        return None

    return TokenConfig(
        name=token_config.name,
        clientId=client_id or token_config.clientId,
        clientSecret=client_secret or token_config.clientSecret,
        apiServer=api_server or token_config.apiServer or "https://api.chalk.ai",
        activeEnvironment=active_environment or token_config.activeEnvironment,
    )



from __future__ import annotations

from datetime import timedelta
from typing import TYPE_CHECKING, Any, Generic, List, Literal, Mapping, Optional, Set, Type, TypeVar, Union, cast

from chalk._validation.feature_validation import FeatureValidation
from chalk.features._encoding.primitive import TPrimitive
from chalk.utils.collections import ensure_tuple
from chalk.utils.duration import Duration, parse_chalk_duration, timedelta_to_duration

T = TypeVar("T")
TPrim = TypeVar("TPrim", bound=TPrimitive)
TRich = TypeVar("TRich")

if TYPE_CHECKING:
    import polars as pl

    from chalk.features._encoding.converter import TDecoder, TEncoder


class WindowedInstance(Generic[T]):
    def __init__(self, values: Mapping[str, T]):
        self.values = values

    def __call__(self, period: str):
        return self.values[period]


class WindowedMeta(type, Generic[T]):
    def __getitem__(cls, underlying: Type[T]) -> "Windowed[Type[T]]":
        return Windowed(
            kind=underlying,
            buckets=[],
            mode="tumbling",
            description=None,
            owner=None,
            tags=None,
            name=None,
            default=None,
            max_staleness=None,
            version=None,
            etl_offline_to_online=None,
            encoder=None,
            decoder=None,
            min=None,
            max=None,
            min_length=None,
            max_length=None,
            contains=None,
            dtype=None,
        )  # noqa


JsonValue = Any


def get_duration_secs(duration: Union[str, int, timedelta]) -> int:
    if isinstance(duration, str):
        duration = parse_chalk_duration(duration)
    if isinstance(duration, timedelta):
        duration_secs_float = duration.total_seconds()
        duration_secs_int = int(duration_secs_float)
        if duration_secs_float != duration_secs_int:
            raise ValueError("Windows that are fractions of seconds are not yet supported")
        duration = duration_secs_int
    return duration


def get_name_with_duration(name_or_fqn: str, duration: Union[str, int, timedelta]) -> str:
    duration_secs = get_duration_secs(duration)
    return f"{name_or_fqn}__{duration_secs}__"


class Windowed(Generic[TPrim, TRich], metaclass=WindowedMeta):
    """Declare a windowed feature.

    Examples
    --------
    >>> @features
    ... class User:
    ...     failed_logins: Windowed[int] = windowed("10m", "24h")
    """

    _mode: Literal["tumbling", "continuous"]

    @property
    def buckets_seconds(self) -> Set[int]:
        return set(int(parse_chalk_duration(bucket).total_seconds()) for bucket in self._buckets)

    @property
    def kind(self) -> Type[TRich]:
        if self._kind is None:
            raise RuntimeError("Window type has not yet been parsed")
        return self._kind

    @kind.setter
    def kind(self, kind: Type[TRich]) -> None:
        assert self._kind is None, "Window type cannot be set twice"
        self._kind = kind

    def _to_feature(self, bucket: Optional[Union[int, str]]):
        from chalk.features import Feature

        assert self._name is not None

        if bucket is None:
            name = self._name
        else:
            if get_duration_secs(bucket) not in self.buckets_seconds:
                raise ValueError(f"Bucket {bucket} is not in the list of specified buckets")
            name = get_name_with_duration(self._name, bucket)

        return Feature(
            name=name,
            version=self._version,
            owner=self._owner,
            tags=None if self._tags is None else list(ensure_tuple(self._tags)),
            description=self._description,
            primary=False,
            default=self._default,
            max_staleness=(
                timedelta_to_duration(self._max_staleness)
                if isinstance(self._max_staleness, timedelta)
                else self._max_staleness
            ),
            etl_offline_to_online=self._etl_offline_to_online,
            encoder=self._encoder,
            decoder=self._decoder,
            pyarrow_dtype=self._dtype,
            validations=FeatureValidation(
                min=self._min,
                max=self._max,
                min_length=self._min_length,
                max_length=self._max_length,
                contains=self._contains,
            ),
            # Only the root feature should have all the durations
            # The pseudofeatures, which are bound to a duration, should not have the durations
            # of the other buckets
            window_durations=tuple(self.buckets_seconds) if bucket is None else tuple(),
            window_duration=None if bucket is None else get_duration_secs(bucket),
            window_mode=self._mode,
        )

    def __init__(
        self,
        buckets: List[str],
        mode: Literal["tumbling", "continuous"],
        description: Optional[str],
        owner: Optional[str],
        tags: Optional[Any],
        name: Optional[str],
        default: Union[TRich, ellipsis],
        max_staleness: Union[Duration, None, ellipsis],
        version: Optional[int],
        etl_offline_to_online: Optional[bool],
        encoder: Optional[TEncoder[TPrim, TRich]],
        decoder: Optional[TDecoder[TPrim, TRich]],
        min: Optional[TRich],
        max: Optional[TRich],
        min_length: Optional[int],
        max_length: Optional[int],
        contains: Optional[TRich],
        dtype: Optional[Union[Type[pl.DataType], pl.DataType]],
        kind: Optional[Type[TRich]],
    ):
        self._kind = kind
        self._name: Optional[str] = None
        self._buckets = buckets
        self._mode = mode
        self._description = description
        self._owner = owner
        self._tags = tags
        self._name = name
        self._default = default
        self._max_staleness = max_staleness
        self._description = description
        self._version = version
        self._etl_offline_to_online = etl_offline_to_online
        self._encoder = encoder
        self._decoder = decoder
        self._min = min
        self._max = max
        self._min_length = min_length
        self._max_length = max_length
        self._contains = contains
        self._dtype = dtype


class SelectedWindow:
    def __init__(self, kind: Windowed, selected: str):
        self.windowed = kind
        self.selected = selected


def windowed(
    *buckets: str,
    description: Optional[str] = None,
    owner: Optional[str] = None,
    tags: Optional[Any] = None,
    name: Optional[str] = None,
    default: Union[T, None, ellipsis] = ...,
    max_staleness: Union[Duration, None, ellipsis] = ...,
    version: Optional[int] = None,
    etl_offline_to_online: Optional[bool] = None,
    encoder: Optional[TEncoder[TPrim, TRich]] = None,
    decoder: Optional[TDecoder[TPrim, TRich]] = None,
    min: Optional[T] = None,
    max: Optional[T] = None,
    min_length: Optional[int] = None,
    max_length: Optional[int] = None,
    contains: Optional[T] = None,
    dtype: Optional[Union[Type[pl.DataType], pl.DataType]] = None,
    # Deprecated
    mode: Literal["tumbling", "continuous"] = "tumbling",
) -> Windowed[TPrim, TRich]:
    """Create a windowed feature.

    See more at https://docs.chalk.ai/docs/aggregations#windowed-features

    Parameters
    ----------
    buckets
        The size of the buckets for the window function.
    default
        The default value of the feature if it otherwise can't be computed.
    owner
        You may also specify which person or group is responsible for a feature.
        The owner tag will be available in Chalk's web portal.
        Alerts that do not otherwise have an owner will be assigned
        to the owner of the monitored feature.
    tags
        Add metadata to a feature for use in filtering, aggregations,
        and visualizations. For example, you can use tags to assign
        features to a team and find all features for a given team.
    max_staleness
        When a feature is expensive or slow to compute, you may wish to cache its value.
        Chalk uses the terminology "maximum staleness" to describe how recently a feature
        value needs to have been computed to be returned without re-running a resolver.

        Read more at https://docs.chalk.ai/docs/feature-caching
    version
        Feature versions allow you to manage a feature as its
        definition changes over time.

        The `version` keyword argument allows you to specify the
        maximum number of versions available for this feature.

        See more at https://docs.chalk.ai/docs/feature-versions
    etl_offline_to_online
        When `True`, Chalk copies this feature into the online environment
        when it is computed in offline resolvers.

        Read more at https://docs.chalk.ai/docs/reverse-etl
    min
        If specified, when this feature is computed, Chalk will check that `x >= min`.
    max
        If specified, when this feature is computed, Chalk will check that `x <= max`.
    min_length
        If specified, when this feature is computed, Chalk will check that `len(x) >= min_length`.
    max_length
        If specified, when this feature is computed, Chalk will check that `len(x) <= max_length`.
    contains

    Other Parameters
    ----------------
    name
        The name for the feature. By default, the name of a feature is
        the name of the attribute on the class, prefixed with
        the camel-cased name of the class. Note that if you provide an
        explicit name, the namespace, determined by the feature class,
        will still be prepended. See `features` for more details.
    description
        Descriptions are typically provided as comments preceding
        the feature definition. For example, you can document a
        `email_count` feature with information about the values
        as follows::
        >>> @features
        ... class User:
        ...     # Count of emails sent
        ...     email_count: Windowed[int] = windowed("10m", "30m")

        You can also specify the description directly with this parameter.
        >>> @features
        ... class User:
        ...     email_count: Windowed[int] = windowed(
        ...         "10m", "30m"
        ...         description="Count of emails sent",
        ...     )
    encoder
    decoder
    dtype
    mode

    Returns
    -------
    Windowed[TPrim, TRich]
        Metadata for the windowed feature, parameterized by
        `TPrim` (the primitive type of the feature) and
        `TRich` (the decoded type of the feature, if `decoder` is provided).
    """
    return Windowed(
        list(buckets),
        mode=mode,
        description=description,
        owner=owner,
        tags=tags,
        name=name,
        default=default,
        max_staleness=max_staleness,
        version=version,
        etl_offline_to_online=etl_offline_to_online,
        encoder=cast("TEncoder", encoder),
        decoder=decoder,
        min=min,
        max=max,
        min_length=min_length,
        max_length=max_length,
        contains=contains,
        dtype=dtype,
        kind=None,
    )



from __future__ import annotations

import enum
from typing import Any, List, Literal, Optional, Union, cast

import pyarrow as pa
from pydantic import BaseConfig, BaseModel, Field
from typing_extensions import Annotated

from chalk.utils.json import TJSON

__all__ = [
    "SerializedDType",
    "BaseSerializedDType",
    "deserialize_dtype_json",
    "serialize_pyarrow_dtype",
]


class DTypeCode(enum.IntEnum):
    # NULL = 1
    BOOL = 2
    INT8 = 3
    INT16 = 4
    INT32 = 5
    INT64 = 6
    UINT8 = 7
    UINT16 = 8
    UINT32 = 9
    UINT64 = 10
    FLOAT16 = 11
    FLOAT32 = 12
    FLOAT64 = 13
    TIME32 = 14  # time unit
    TIME64 = 15  # time unit
    TIMESTAMP = 16  # time unit, time zone
    DATE32 = 17
    DATE64 = 18
    DURATION = 19  # time unit
    BINARY = 20  # length
    STRING = 21
    LARGE_BINARY = 22
    LARGE_STRING = 23
    # Decimal is not yet supported by chalk
    # DECIMAL128 = 24  # precision, scale
    LIST = 25  # value type, size
    LARGE_LIST = 26  # value type
    STRUCT = 27  # fields


class TimeUnitCode(str, enum.Enum):
    SECOND = "s"
    MILLISECONDS = "ms"
    MICROSECONDS = "us"
    NANOSECONDS = "ns"


class BaseSerializedDType(BaseModel):
    type_code: DTypeCode

    def to_pyarrow_dtype(self) -> pa.DataType:
        raise NotImplementedError

    class Config(BaseConfig):
        use_enum_values = True


class SingletonDType(BaseSerializedDType):
    type_code: Literal[
        DTypeCode.BOOL,
        DTypeCode.INT8,
        DTypeCode.INT16,
        DTypeCode.INT32,
        DTypeCode.INT64,
        DTypeCode.UINT8,
        DTypeCode.UINT16,
        DTypeCode.UINT32,
        DTypeCode.UINT64,
        DTypeCode.FLOAT16,
        DTypeCode.FLOAT32,
        DTypeCode.FLOAT64,
        DTypeCode.DATE32,
        DTypeCode.DATE64,
        DTypeCode.STRING,
        DTypeCode.LARGE_STRING,
        DTypeCode.LARGE_BINARY,
    ]

    def to_pyarrow_dtype(self) -> pa.DataType:
        if self.type_code == DTypeCode.BOOL:
            return pa.bool_()
        try:
            return getattr(pa, self.type_code.name.lower())()
        except AttributeError:
            raise ValueError(f"Unsupported dtype: {self.type_code}") from None


class TimeUnitDType(BaseSerializedDType):
    type_code: Literal[DTypeCode.TIME32, DTypeCode.TIME64, DTypeCode.DURATION]
    time_unit: TimeUnitCode

    def to_pyarrow_dtype(self) -> pa.DataType:
        return getattr(pa, self.type_code.name.lower())(self.time_unit)

    class Config(BaseSerializedDType.Config):
        use_enum_values = True


class TimestampDType(BaseSerializedDType):
    type_code: Literal[DTypeCode.TIMESTAMP]
    time_unit: TimeUnitCode
    timezone: Optional[str]

    def to_pyarrow_dtype(self) -> pa.DataType:
        return pa.timestamp(self.time_unit, self.timezone)


class BinaryDType(BaseSerializedDType):
    type_code: Literal[DTypeCode.BINARY]
    length: int

    def to_pyarrow_dtype(self) -> pa.DataType:
        return pa.binary(self.length)


class ListDType(BaseSerializedDType):
    type_code: Literal[DTypeCode.LIST]
    inner_dtype: BaseSerializedDType
    length: int

    def to_pyarrow_dtype(self) -> pa.DataType:
        return pa.list_(self.inner_dtype.to_pyarrow_dtype(), self.length)


class LargeListDType(BaseSerializedDType):
    type_code: Literal[DTypeCode.LARGE_LIST]
    inner_dtype: BaseSerializedDType

    def to_pyarrow_dtype(self) -> pa.DataType:
        return pa.large_list(self.inner_dtype.to_pyarrow_dtype())


class StructField(BaseModel):
    name: str
    dtype: BaseSerializedDType
    nullable: bool = True

    def to_pyarrow_dtype(self) -> pa.Field:
        return pa.field(self.name, self.dtype.to_pyarrow_dtype(), self.nullable)


class StructDType(BaseSerializedDType):
    type_code: Literal[DTypeCode.STRUCT]
    fields: List[StructField]

    def to_pyarrow_dtype(self) -> pa.Field:
        return pa.struct([x.to_pyarrow_dtype() for x in self.fields])


SerializedDType = Annotated[
    Union[SingletonDType, TimeUnitDType, TimestampDType, BinaryDType, ListDType, LargeListDType, StructDType],
    Field(discriminator="type_code"),
]


class _Deserializer(BaseModel):
    __root__: SerializedDType


def deserialize_dtype_json(dtype_json: TJSON) -> SerializedDType:
    deserializer = _Deserializer(__root__=cast(Any, dtype_json))
    return deserializer.__root__


def serialize_pyarrow_dtype(pa_dtype: pa.DataType) -> SerializedDType:
    if pa.types.is_boolean(pa_dtype):
        return SingletonDType(type_code=DTypeCode.BOOL)
    if pa.types.is_uint8(pa_dtype):
        return SingletonDType(type_code=DTypeCode.UINT8)
    if pa.types.is_uint16(pa_dtype):
        return SingletonDType(type_code=DTypeCode.UINT16)
    if pa.types.is_uint32(pa_dtype):
        return SingletonDType(type_code=DTypeCode.UINT32)
    if pa.types.is_uint64(pa_dtype):
        return SingletonDType(type_code=DTypeCode.UINT64)
    if pa.types.is_int8(pa_dtype):
        return SingletonDType(type_code=DTypeCode.INT8)
    if pa.types.is_int16(pa_dtype):
        return SingletonDType(type_code=DTypeCode.INT16)
    if pa.types.is_int32(pa_dtype):
        return SingletonDType(type_code=DTypeCode.INT32)
    if pa.types.is_int64(pa_dtype):
        return SingletonDType(type_code=DTypeCode.INT64)
    if pa.types.is_float16(pa_dtype):
        return SingletonDType(type_code=DTypeCode.FLOAT16)
    if pa.types.is_float32(pa_dtype):
        return SingletonDType(type_code=DTypeCode.FLOAT32)
    if pa.types.is_float64(pa_dtype):
        return SingletonDType(type_code=DTypeCode.FLOAT64)
    if pa.types.is_time32(pa_dtype):
        return TimeUnitDType(type_code=DTypeCode.TIME32, time_unit=TimeUnitCode(pa_dtype.unit))

    if pa.types.is_time64(pa_dtype):
        return TimeUnitDType(type_code=DTypeCode.TIME64, time_unit=TimeUnitCode(pa_dtype.unit))

    if pa.types.is_timestamp(pa_dtype):
        return TimestampDType(
            type_code=DTypeCode.TIMESTAMP,
            time_unit=TimeUnitCode(pa_dtype.unit),
            timezone=pa_dtype.tz,
        )

    if pa.types.is_date32(pa_dtype):
        return SingletonDType(type_code=DTypeCode.DATE32)
    if pa.types.is_date64(pa_dtype):
        return SingletonDType(type_code=DTypeCode.DATE64)
    if pa.types.is_duration(pa_dtype):
        return TimeUnitDType(type_code=DTypeCode.DURATION, time_unit=TimeUnitCode(pa_dtype.unit))

    if pa.types.is_large_string(pa_dtype):
        return SingletonDType(type_code=DTypeCode.LARGE_STRING)
    if pa.types.is_large_binary(pa_dtype):
        return SingletonDType(type_code=DTypeCode.LARGE_BINARY)
    if pa.types.is_fixed_size_binary(pa_dtype):
        return BinaryDType(type_code=DTypeCode.BINARY, length=pa_dtype.byte_width)
    if pa.types.is_binary(pa_dtype):
        return BinaryDType(type_code=DTypeCode.BINARY, length=-1)
    if pa.types.is_string(pa_dtype):
        return SingletonDType(type_code=DTypeCode.STRING)

    if pa.types.is_large_list(pa_dtype):
        return LargeListDType(type_code=DTypeCode.LARGE_LIST, inner_dtype=serialize_pyarrow_dtype(pa_dtype.value_type))
    if pa.types.is_fixed_size_list(pa_dtype):
        return ListDType(
            type_code=DTypeCode.LIST,
            inner_dtype=serialize_pyarrow_dtype(pa_dtype.value_type),
            length=pa_dtype.list_size,
        )

    if pa.types.is_list(pa_dtype):
        return ListDType(
            type_code=DTypeCode.LIST,
            inner_dtype=serialize_pyarrow_dtype(pa_dtype.value_type),
            length=-1,
        )

    if pa.types.is_struct(pa_dtype):
        schema = pa.schema(pa_dtype)
        return StructDType(
            type_code=DTypeCode.STRUCT,
            fields=[
                StructField(name=name, dtype=serialize_pyarrow_dtype(dtype))
                for (name, dtype) in zip(schema.names, schema.types)
            ],
        )

    raise ValueError(f"Unsupported dtype: {pa_dtype}")



from __future__ import annotations

from os import PathLike
from typing import TYPE_CHECKING, Any, Dict, Optional, TypeVar, Union

from chalk.sql._internal.sql_source import BaseSQLSource, TableIngestMixIn
from chalk.sql.protocols import SQLSourceWithTableIngestProtocol
from chalk.utils.missing_dependency import missing_dependency_exception

if TYPE_CHECKING:
    from sqlalchemy.engine.url import URL


T = TypeVar("T")


class SQLiteInMemorySourceImpl(TableIngestMixIn, BaseSQLSource, SQLSourceWithTableIngestProtocol):
    def __init__(self, name: Optional[str] = None):
        self.ingested_tables: Dict[str, Any] = {}
        BaseSQLSource.__init__(self, name=name)

    def local_engine_url(self) -> URL:
        try:
            from sqlalchemy.engine.url import URL
        except ImportError:
            raise missing_dependency_exception("chalkpy[sqlite]")
        return URL.create(drivername="sqlite", database=":memory:", query={"check_same_thread": "true"})


class SQLiteFileSourceImpl(TableIngestMixIn, BaseSQLSource, SQLSourceWithTableIngestProtocol):
    def __init__(self, filename: Union[PathLike, str], name: Optional[str] = None):
        self.filename = str(filename)
        self.ingested_tables: Dict[str, Any] = {}
        BaseSQLSource.__init__(self, name=name)

    def local_engine_url(self) -> URL:
        try:
            from sqlalchemy.engine.url import URL
        except ImportError:
            raise missing_dependency_exception("chalkpy[sqlite]")
        return URL.create(drivername="sqlite", database=self.filename, query={"check_same_thread": "true"})



import abc
from typing import List, Optional

from chalk.client.models import ChalkError


class ChalkBaseException(Exception, abc.ABC):
    """The base type for Chalk exceptions.

    This exception makes error handling easier, as you can
    look only for this exception class.
    """

    message: str
    """A message describing the specific type of exception raised."""

    full_message: str
    """A message that describes the specific type of exception raised
    and contains the readable representation of each error in the
    errors attribute.
    """

    errors: List[ChalkError]
    """The errors from executing a Chalk operation.

    These errors contain more detailed information about
    why the exception occurred.
    """

    def __init__(self, errors: Optional[List[ChalkError]] = None):
        if errors is None:
            errors = []

        self.errors = errors

        super().__init__(self.full_message)

    @property
    def message(self) -> str:
        raise NotImplementedError

    @property
    def full_message(self) -> str:
        if self.errors:
            return self.message + "\n" + "\n".join(["\t" + e.message for e in self.errors])

        return self.message


class ChalkWhoAmIException(ChalkBaseException):
    """Raised from the `ChalkClient.whoami` method."""

    @property
    def message(self) -> str:
        return "Failed to retrieve current user information during `whoami` check"


class ChalkOnlineQueryException(ChalkBaseException):
    @property
    def message(self) -> str:
        return "Failed to execute online query"


class ChalkOfflineQueryException(ChalkBaseException):
    @property
    def message(self) -> str:
        return "Failed to execute offline query"

    @property
    def full_message(self) -> str:
        return self.message + "\n" + "\n".join(["\t" + e.message for e in self.errors[0:3]])


class ChalkComputeResolverException(ChalkOfflineQueryException):
    @property
    def message(self) -> str:
        return "Failed to compute resolver output"


class ChalkResolverRunException(ChalkBaseException):
    """Raised when failing to get the resolver's status via `ChalkClient.get_run_status`"""

    @property
    def message(self) -> str:
        return "Resolver run failed"


class ChalkDatasetDownloadException(ChalkBaseException):
    @property
    def message(self) -> str:
        return "Failed to download dataset"


class ChalkInvalidCredentialsException(ChalkBaseException):
    """Raised when constructing a `ChalkClient` without valid credentials.

    When this exception is raise, no explicit `client_id` and `client_secret`
    were provided, there was no `~/.chalk.yml` file with applicable credentials,
    and the environment variables `CHALK_CLIENT_ID` and `CHALK_CLIENT_SECRET`
    were not set.

    You may need to run `chalk login` from your command line, or check that your
    working directory is set to the root of your project.
    """

    message = (
        "Explicit `client_id` and `client_secret` are not provided, "
        "there is no `~/.chalk.yml` file with applicable credentials, "
        "and the environment variables `CHALK_CLIENT_ID` and "
        "`CHALK_CLIENT_SECRET` are not set. "
        "You may need to run `chalk login` from your command line, "
        "or check that your working directory is set to the root of "
        "your project."
    )


__all__ = [
    "ChalkBaseException",
    "ChalkOnlineQueryException",
    "ChalkOfflineQueryException",
    "ChalkResolverRunException",
    "ChalkDatasetDownloadException",
    "ChalkComputeResolverException",
    "ChalkInvalidCredentialsException",
    "ChalkWhoAmIException",
]
